/* kbdsniffd - keyboard-sniffer-daemon
 * Author: Marco Skambraks <marco@suse.de>
*/

#include <getopt.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <pthread.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/kd.h>
#include <linux/keyboard.h>
#include <linux/vt.h>
#include "uinput.h"
#include "ringbuffer.h"
#include "event.h"
#include "load_config.h"

#define EVENT_TIMEOUT 200       /* event timeout in ms */
#define CONN_TIMEOUT 2          /* connection timeeout in seconds */
#define MAX_COUNT 256           /* maximum events during one read */
#define MAX_DEV_NUM 23
#define KEY_OFFSET 500          /* offset to indicate special keys */
#define SOCK_NAME "/var/run/kbd.sock"
#define CONSOLE_DEV "/dev/console"
#define MAX_STR 255

#define USAGE  " \
kbdsniffd version 0.6 \n \
keyboard sniffer daemon - to by-pass keyboard events \n \
Author: Marco Skambraks <marco@suse.de> \n\n \
-n --nodaemon - don't fork\n \
-s --scan - scan for first keyboard device number \n \
-h --help - display this help \n \
-b --basedir - device basedir (/dev/input) \n \
-c --config - config file (/etc/kbdsniffd.conf) \n \
-u --uinput - uinput device (uinput) \n \
-p --port - port for incoming connections - 0 means unix socket (0)\n \
-k --key - authentication key (default) \n \
-v --verbose - be verbose \n\n"

struct option long_options[] = {
  {"nodaemon", 0, 0, 'n'},
  {"verbose", 1, 0, 'v'},
  {"basedir", 1, 0, 'b'},
  {"uinput", 1, 0, 'u'},
  {"help", 0, 0, 'h'},
  {"port", 1, 0, 'p'},
  {"key", 1, 0, 'k'},
  {"config", 1, 0, 'c'},
  {"scan", 0, 0, 's'},
  {0, 0, 0, 0}
};

void *grabber_thread (void *arg);

int keycode_trans (int keycode);

void close_all ();

/* handler for signal "sigpipe" */
void sigpipe (int signum);

/* termination handler */
void terminate (int signum);

/* socket communication */
void handle_connection (int sock_id);

/* helper function to send message over the socket */
int sock_send (int sock_id, char *str);

/* helper function - parse the config file */
int load_config (char *file);

void kbd_scan ();

int elapsed_sec (struct timeval *start);

/* global variables */
int verbose = 0;

char authkey[MAX_STR] = "default";

int port = 0;

char basedir[MAX_STR] = "/dev/input";

char uinput_dev[MAX_STR] = "uinput";

int autoscan = 0;

int sock_listen = 0;

int con_fd = 0;

volatile sig_atomic_t grabber_loop = 1;

/* array for the connected keyboards */
struct kbd_prop
{
  int dev_num;
  int on;
  pthread_t id;
  char *basedir;
} kbd_connected[MAX_DEV_NUM];

volatile sig_atomic_t main_loop = 1;

volatile sig_atomic_t conn_loop = 1;

volatile int sniffer_key = 0;

pthread_mutex_t key_buf_mutex = PTHREAD_MUTEX_INITIALIZER;

pthread_mutex_t loop_mutex = PTHREAD_MUTEX_INITIALIZER;

int main (int argc, char **argv)
{

  int c = 0;

  int do_fork = 1;

  struct hostent *hp = NULL, *gethostbyname ();

  int tmp = 0;

  char conf[MAX_STR] = "/etc/kbdsniffd.conf";

  struct sockaddr_in sock_addr;

  struct sockaddr_un sock_file;

  int sock_new = 0;

  int sock_reuse = 1;

  int sock_ret = 0;

  memset (&kbd_connected, 0, sizeof (kbd_connected));
  /* try to load config from default path */
  load_config (conf);

  /* parse options */
  while (1)
   {
     int option_index = 0;

     c =
       getopt_long (argc, argv, "nsb:c:hvk:u:p:", long_options,
                    &option_index);
     if (c == -1)
       break;

     switch (c)
      {
      case 0:
        fprintf (stderr, " option %s", long_options[option_index].name);
        if (optarg)
          fprintf (stderr, " with arg %s", optarg);
        fprintf (stderr, "\n");
        break;

      case 'h':
        fprintf (stderr, "%s", USAGE);
        return 0;

      case 'b':
        strcpy (basedir, optarg);
        break;
      case 'p':
        port = atoi (optarg);
        break;
      case 'k':
        strcpy (authkey, optarg);
        break;
      case 'u':
        strcpy (uinput_dev, optarg);
        break;
      case 'c':
        strcpy (conf, optarg);
        load_config (conf);
        break;
      case 'v':
        verbose = 1;
        break;

      case 's':
       {
         do
          {
            tmp = event_scan (basedir, tmp, MAX_DEV_NUM);
            if (tmp >= 0)
             {
               fprintf (stderr, "kbd=%d \n", tmp);
               tmp++;
             }

          }
         while (tmp >= 0);
       }
        return 0;

      case 'n':
        do_fork = 0;
        break;

      case '?':
        return 2;
        break;

      }
   }

  if (optind < argc)
   {
     fprintf (stderr, " non - option: ");
     return 2;
     while (optind < argc)
       fprintf (stderr, "%s", argv[optind++]);
     fprintf (stderr, "\n");

     return 2;
   }

  if (do_fork)
   {

     switch (fork ())
      {
      case -1:
        fprintf (stderr, "error: fork failed \n");
        return 1;
      case 0:
        setsid ();
        /* become a daemon */
        break;
      default:
        return 0;
      }

   }

  if (verbose)
    fprintf (stderr, "basedir=%s uinput=%s - localhost:%d\n", basedir,
             uinput_dev, port);

  if (uinput_init (basedir, uinput_dev))
   {
     fprintf (stderr, "%s", USAGE);
     fprintf (stderr,
              "can't open %s/%s \nplease make sure that the module uinput is loaded\n",
              basedir, uinput_dev);
     return 1;
   }

  if ((con_fd = open (CONSOLE_DEV, O_RDWR)) < 0)
   {
     close_all ();
     fprintf (stderr, "can't open %s\n", CONSOLE_DEV);
     return 3;
   }

  /* set socket properties */
  if (!port)                    /* use uunix socket */
    sock_listen = socket (AF_UNIX, SOCK_STREAM, 0);
  else                          /* use inet socket */
    sock_listen = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP);

  sock_ret =
    setsockopt (sock_listen, SOL_SOCKET, SO_REUSEADDR, &sock_reuse,
                sizeof (sock_reuse));
  if (sock_ret < 0)
   {
     close_all ();
     fprintf (stderr, "can't set socket option\n");
     return 4;
   }

  if (!port)                    /* unix socket */
   {
     sock_file.sun_family = AF_UNIX;
     strcpy (sock_file.sun_path, SOCK_NAME);
     unlink (SOCK_NAME);

     sock_ret =
       bind (sock_listen, (struct sockaddr *) &sock_file, sizeof (sock_file));
   }
  else
   {
     sock_addr.sin_family = AF_INET;
     sock_addr.sin_port = htons (port);

     if ((hp = gethostbyname ("localhost")) == NULL)
      {
        close_all ();
        fprintf (stderr, "can't get host (localhost)\n");
        return 5;
      }
     else
       memcpy (&sock_addr.sin_addr, hp->h_addr, hp->h_length);

     sock_ret =
       bind (sock_listen, (struct sockaddr *) &sock_addr, sizeof (sock_addr));
   }                            /* else - inet socket */

  if (sock_ret < 0)
   {
     close_all ();
     fprintf (stderr, "can't bind socket\n");
     return 7;
   }

  if (listen (sock_listen, 0) < 0)
   {
     close_all ();
     fprintf (stderr, "liten failed\n");
     return 8;
   }

  signal (SIGPIPE, sigpipe);
  signal (SIGINT, terminate);
  signal (SIGTERM, terminate);
  signal (SIGCHLD, SIG_IGN);
  signal (SIGHUP, SIG_IGN);

  while (main_loop)
   {
     int threads_running = 1;

     if (verbose)
       fprintf (stderr, "listening on port=%d\n", port);

     sock_new = accept (sock_listen, NULL, NULL);
     if (sock_new >= 0)
      {

        grabber_loop = 1;
        conn_loop = 1;
        key_buf_init ();

        handle_connection (sock_new);
        if (sock_new >= 0)
          close (sock_new);

        pthread_mutex_lock (&loop_mutex);
        grabber_loop = 0;
        pthread_mutex_unlock (&loop_mutex);

        while (threads_running)
         {
           int i;

           threads_running = 0;
           for (i = 0; i < MAX_DEV_NUM; i++)
             if (kbd_connected[i].basedir)
               threads_running++;
         }                      /* while threads */
      }

   }

  if (sock_listen >= 0)
    close (sock_listen);
  close_all ();
  if (!port)
    unlink (SOCK_NAME);

  if (verbose)
    fprintf (stderr, "%s finished\n", argv[0]);

  return 0;
}

int keycode_trans (int keycode)
{
  struct kbentry key;

  int type = 0;

  switch (keycode)
   {
   case 1:
   case 14:
   case 15:
   case 28:
   case 57:
     return (keycode + KEY_OFFSET);
   }

  key.kb_table = 0;
  key.kb_index = keycode;
  key.kb_value = 0;
  if (ioctl (con_fd, KDGKBENT, &key) < 0)
    return 0;

  type = KTYP (key.kb_value) & 0x0f;

  if (type == KT_LETTER || type == KT_LATIN)
    return KVAL (key.kb_value);
  else
    return (keycode + KEY_OFFSET);
}

void *grabber_thread (void *arg)
{
  int i = 0, rd = 0;

  struct kbd_prop *kbd = (struct kbd_prop *) arg;

  int sniff_on = 0;

  struct input_event ev[MAX_COUNT];

  int con_fd;                   /* console fd */

  int con_nr = 0;               /* current console */

  int con_mode = 0;             /* console mode 0 = text - other values
                                   indicates none-text-mode */
  struct vt_stat vterm;

  int event_err = 0;

  int event_fd = 0;

  int do_exit = 0;

  if ((event_fd = event_init (kbd->basedir, kbd->dev_num)) < 0)
   {
     if (verbose)
       fprintf (stderr, "can't open %s/event%d\n", kbd->basedir,
                kbd->dev_num);
     kbd->basedir = 0;
     kbd->on = 0;
     kbd->id = 0;
     kbd->dev_num = 0;
     return 0;
   }
  if (verbose)
    fprintf (stderr, "grabber thread started pid=%d\n",
             (unsigned int) getpid ());

  /* wait until a sniffer key is set or grabber is disabled */
  while (grabber_loop && !sniffer_key)
    usleep (1000);

  if (verbose)
    fprintf (stderr, "switch to grabbing mode - key=%d\n", sniffer_key);

  /* do a dummy read to drop buffered keys */
  while (event_get (event_fd, ev, MAX_COUNT, 100) != -2)
    usleep (1000);

  event_grab (event_fd, 1);

  con_fd = open ("/dev/console", O_RDWR);
  while (grabber_loop && !do_exit)
   {

     rd = event_get (event_fd, ev, MAX_COUNT, EVENT_TIMEOUT);
     if (rd == -3)
       event_err = 1;

     for (i = 0; i < rd; i++)
      {

        if (ev[i].type == 1 && ev[i].code == sniffer_key && ev[i].value > 0
            && !con_mode)
          sniff_on = 1;

        /* if not in sniffer mode send the keys back to the system otherwise
           write the keys to a buffer */

        if (!sniff_on)
         {

           /* check if left-ctrl or right-ctrl is pressed */
           if (ev[i].type == 1 && ev[i].value == 1
               && (ev[i].code == 29 || ev[i].code == 97))
            {
              pthread_mutex_lock (&key_buf_mutex);
              key_buf_init ();
              key_put_buf (999);
              pthread_mutex_unlock (&key_buf_mutex);
            }

           uinput_key (ev[i]);
         }
        else if (ev[i].type == 1 && ev[i].code != sniffer_key && ev[i].value)
         {
           pthread_mutex_lock (&key_buf_mutex);
           key_put_buf (keycode_trans (ev[i].code));
           pthread_mutex_unlock (&key_buf_mutex);
         }

        /* disable sniffer mode if the sniffer key is released */
        if (ev[i].type == 1 && ev[i].code == sniffer_key && !ev[i].value)
          sniff_on = 0;
      }

     if (!sniff_on)
      {
        /* check if console has changed */
        ioctl (con_fd, VT_GETSTATE, &vterm);
        if (vterm.v_active != con_nr)
         {
           int mode_fd;

           char tmpstr[MAX_STR];

           sprintf (tmpstr, "/dev/tty%d", vterm.v_active);
           mode_fd = open (tmpstr, O_RDWR);
           ioctl (mode_fd, KDGETMODE, &con_mode);
           close (mode_fd);
           con_nr = vterm.v_active;
         }                      /* if - changed */

      }

     /* check for event-errors - if true = fallback to the first dev */
     if (event_err)
      {
        event_close (event_fd);
        if (verbose)
          fprintf (stderr, "event error - exit\n");
        do_exit = 1;
      }                         /* if error */
   }

  if (con_fd >= 0)
    close (con_fd);

  if (event_fd >= 0)
    event_close (event_fd);

  kbd->dev_num = 0;
  kbd->on = 0;
  kbd->id = 0;
  kbd->basedir = 0;
  if (verbose)
    fprintf (stderr, "grabber thread finished pid=%d\n",
             (unsigned int) getpid ());

  return 0;
}

void close_all ()
{

  grabber_loop = 0;
  uinput_close ();
  if (con_fd >= 0)
    close (con_fd);
  sleep (1);
}

void handle_connection (int sock_id)
{
  struct timeval timeout;

  int auth = 0;

  fd_set select_set;

  char tmpstr[MAX_STR] = "";

  if (verbose)
    fprintf (stderr, "entering connection loop\n");

  while (conn_loop)
   {
     int ret = 0;

     struct timeval as_timer;   /* kbd autoscan timer */

     FD_ZERO (&select_set);
     FD_SET (sock_id, &select_set);
     timeout.tv_sec = CONN_TIMEOUT;
     timeout.tv_usec = 0;

     memset (tmpstr, 0, sizeof (tmpstr));

     if (auth && autoscan)
      {
        if (elapsed_sec (&as_timer) > autoscan)
         {
           kbd_scan ();
           gettimeofday (&as_timer, NULL);
         }
      }

     ret = select (sock_id + 1, &select_set, NULL, NULL, &timeout);
     if (ret > 0)
      {
        ret = recv (sock_id, tmpstr, MAX_STR, MSG_DONTWAIT);

        if (ret <= 0)
         {
           pthread_mutex_lock (&loop_mutex);
           conn_loop = 0;
           grabber_loop = 0;
           auth = 0;
           pthread_mutex_unlock (&loop_mutex);
           continue;
         }
        /* check string */
        if (!strncmp (tmpstr, "auth ", 5))
         {
           tmpstr[strlen (tmpstr) - 2] = 0;
           if (!strcmp (tmpstr + 5, authkey))
            {
              auth = 1;
              conn_loop = sock_send (sock_id, "OK\r\n");
              kbd_scan ();
              gettimeofday (&as_timer, NULL);
            }
           else
            {
              sock_send (sock_id, "bye\r\n");
              conn_loop = 0;
              auth = 0;
            }

         }
        else if (!strcmp (tmpstr, "get\r\n") && auth)
         {
           char key_str[10] = "";

           int key = 0;

           if (key_getbufcnt ())
            {
              pthread_mutex_lock (&key_buf_mutex);
              key_get_buf (&key);
              pthread_mutex_unlock (&key_buf_mutex);
              sprintf (key_str, "%03d\r\n", key);
            }
           else
             strcpy (key_str, "000\r\n");

           conn_loop = sock_send (sock_id, key_str);
         }
        else if (!strncmp (tmpstr, "setkey ", 7) && auth)
         {
           int key = 0;

           key = atoi (tmpstr + 7);
           if (key > KEY_OFFSET)
            {
              pthread_mutex_lock (&loop_mutex);
              sniffer_key = key - KEY_OFFSET;
              conn_loop = sock_send (sock_id, "OK\r\n");
              pthread_mutex_unlock (&loop_mutex);

            }
           else
             conn_loop =
               sock_send (sock_id, "error: this is not a valid key\r\n");

         }

        else if (!strcmp (tmpstr, "quit\r\n"))
         {
           pthread_mutex_lock (&loop_mutex);
           conn_loop = 0;
           grabber_loop = 0;
           auth = 0;
           pthread_mutex_unlock (&loop_mutex);
         }
        else if (!auth)
          conn_loop = sock_send (sock_id, "error: permission denied\r\n");

      }
     else
      {
        sock_send (sock_id, "bye\r\n");
        pthread_mutex_lock (&loop_mutex);
        conn_loop = 0;
        grabber_loop = 0;
        auth = 0;
        pthread_mutex_unlock (&loop_mutex);
      }

   }

  if (verbose)
    fprintf (stderr, "connection finished\n");

}

int sock_send (int sock_id, char *str)
{
  int ret = 0;

  fd_set select_set;

  FD_ZERO (&select_set);
  FD_SET (sock_id, &select_set);
  struct timeval timeout;

  timeout.tv_sec = 1;
  timeout.tv_usec = 0;
  if (select (sock_id + 1, NULL, &select_set, NULL, &timeout) <= 0)
    return 0;

/* send the string including the \0 character */
  ret = send (sock_id, str, strlen (str), MSG_DONTWAIT);

  if (ret != (int) (strlen (str)))
   {
     if (verbose)
       fprintf (stderr, "sock_send: %d\n", ret);

     return 0;
   }

  return 1;
}

void sigpipe (int signum)
{
  pthread_mutex_lock (&loop_mutex);
  conn_loop = 0;
  grabber_loop = 0;
  pthread_mutex_unlock (&loop_mutex);
  signal (signum, sigpipe);
}

void terminate (int signum)
{
  pthread_mutex_lock (&loop_mutex);
  conn_loop = 0;
  grabber_loop = 0;
  main_loop = 0;
  pthread_mutex_unlock (&loop_mutex);
  fcntl (sock_listen, F_SETFL, O_NONBLOCK);
  signal (signum, terminate);
}

int load_config (char *file)
{
  char tmpstr[MAX_STR] = "";

  if (confread (file))
    return 1;

  confsearch (strcpy (tmpstr, "key"));
  if (tmpstr[0])
    strcpy (authkey, tmpstr);

  memset (tmpstr, 0, MAX_STR);

  confsearch (strcpy (tmpstr, "port"));
  if (atoi (tmpstr) > 0)
    port = atoi (tmpstr);

  memset (tmpstr, 0, MAX_STR);

  confsearch (strcpy (tmpstr, "basedir"));
  if (tmpstr[0])
    strcpy (basedir, tmpstr);

  memset (tmpstr, 0, MAX_STR);

  confsearch (strcpy (tmpstr, "uinput"));
  if (tmpstr[0])
    strcpy (uinput_dev, tmpstr);

  memset (tmpstr, 0, MAX_STR);
  confsearch (strcpy (tmpstr, "autoscan"));
  if (!tmpstr[0])
    autoscan = 0;
  else
    autoscan = atoi (tmpstr);

  return 0;
}

void kbd_scan ()
{
  int tmp = 0;

  /* fill the keyboard array */
  do
   {
     tmp = event_scan (basedir, tmp, MAX_DEV_NUM);
     if (tmp >= 0)
      {
        if (!kbd_connected[tmp].on)
         {

           kbd_connected[tmp].dev_num = tmp;
           kbd_connected[tmp].on = 1;   /* connected but */
           kbd_connected[tmp].basedir = basedir;
           /* start grabber thread */
           pthread_create (&kbd_connected[tmp].id, NULL, grabber_thread,
                           (void *) &kbd_connected[tmp]);
           pthread_detach (kbd_connected[tmp].id);
           sched_yield ();
         }
        tmp++;
      }

   }
  while (tmp >= 0);
}

int elapsed_sec (struct timeval *start)
{
  struct timeval current_time;

  gettimeofday (&current_time, NULL);
  return (current_time.tv_sec - start->tv_sec);
}
